//havok_config.cpp
#pragma comment(lib, "hkBase.lib")
#pragma comment(lib, "hkSerialize.lib")
#pragma comment(lib, "hkSceneData.lib")
#pragma comment(lib, "hkInternal.lib")
#pragma comment(lib, "hkGeometryUtilities.lib")
#pragma comment(lib, "hkVisualize.lib")
#pragma comment(lib, "hkCompat.lib")
#pragma comment(lib, "hkpCollide.lib")
#pragma comment(lib, "hkpconstraintsolver.lib")
#pragma comment(lib, "hkpDynamics.lib")
#pragma comment(lib, "hkpinternal.lib")
#pragma comment(lib, "hkputilities.lib")
#pragma comment(lib, "hkpVehicle.lib")


// [id=keycode]
#include <Common/Base/keycode.cxx>

// [id=productfeatures]
// We're using only physics - we undef products even if the keycode is present so
// that we don't get the usual initialization for these products.
#undef HK_FEATURE_PRODUCT_AI
#undef HK_FEATURE_PRODUCT_ANIMATION
#undef HK_FEATURE_PRODUCT_CLOTH
#undef HK_FEATURE_PRODUCT_DESTRUCTION
#undef HK_FEATURE_PRODUCT_BEHAVIOR
#define HK_FEATURE_REFLECTION_PHYSICS
#define HK_CLASSES_FILE <Common/Serialize/Classlist/hkClasses.h>

// Also we're not using any of these.
//#define HK_EXCLUDE_FEATURE_SerializeDeprecatedPre700
//#define HK_EXCLUDE_FEATURE_RegisterVersionPatches
	//#define HK_EXCLUDE_FEATURE_RegisterReflectedClasses
//#define HK_EXCLUDE_FEATURE_MemoryTracker
//#define HK_EXCLUDE_FEATURE_hkpAccurateInertiaTensorComputer
//#define HK_EXCLUDE_FEATURE_CompoundShape
//#define HK_EXCLUDE_FEATURE_hkpAabbTreeWorldManager
//#define HK_EXCLUDE_FEATURE_hkpContinuousSimulation
//#define HK_EXCLUDE_FEATURE_hkpKdTreeWorldManager


#include <Common/Base/Config/hkProductFeatures.cxx>

#include "havok_config.h"

//HAvok  Ͼ  ȣѴ.

hkpWorld *g_world = NULL;
const float g_groundHalfHeight = 2.0f;
hkpRigidBody* g_groundBody = NULL;

//ڽ    
const float g_bx = 1.0f, g_by = 2.0f, g_bz = 1.0f;
const int g_nx = 2, g_ny = 2, g_nz = 2;
const int g_NUM_BODIES = g_nx * g_ny * g_nz;
hkpRigidBody *g_bodies[g_NUM_BODIES];


void HavokErrorReportFunction(const char* msg, void* errorReportObject) 
{
	OutputDebugString( msg );
	OutputDebugString( "\n" );
}


#define USE_VDB   1

hkJobThreadPool*	g_threadPool = NULL;
int					g_totalNumThreadsUsed = 0;
hkVisualDebugger*	g_vdb = NULL;
hkArray<hkProcessContext*> g_contexts;
hkJobQueue*			g_jobQueue = NULL;

hkStopwatch			g_stopWatch;
hkReal				g_lastTime = 0;

void InitVisualDebuger()
{
#ifdef USE_VDB
	hkpWorld* physicsWorld = g_world;
	hkArray<hkProcessContext*>& contexts = g_contexts;

	// <PHYSICS-ONLY>: Register physics specific visual debugger processes
	// By default the VDB will show debug points and lines, however some products such as physics and cloth have additional viewers
	// that can show geometries etc and can be enabled and disabled by the VDB app.
	hkpPhysicsContext* context;
	{
		// The visual debugger so we can connect remotely to the simulation
		// The context must exist beyond the use of the VDB instance, and you can make
		// whatever contexts you like for your own viewer types.
		context = new hkpPhysicsContext();
		hkpPhysicsContext::registerAllPhysicsProcesses(); // all the physics viewers
		context->addWorld(physicsWorld); // add the physics world so the viewers can see it
		contexts.pushBack(context);

		// Now we have finished modifying the world, release our write marker.
		physicsWorld->unmarkForWrite();
	}

	hkVisualDebugger* vdb = new hkVisualDebugger(contexts);
	vdb->serve();
	g_vdb = vdb;
#else
	g_world->unmarkForWrite();
#endif
}

void UnInitVisualDebuger()
{
#ifdef USE_VDB
	g_world->lock();
	hkArray<hkProcessContext*>& contexts = g_contexts;
	hkVisualDebugger* vdb = g_vdb;

	vdb->shutdown();
	vdb->removeReference();
	vdb = HK_NULL;

	// Contexts are not reference counted at the base class level by the VDB as
	// they are just interfaces really. So only delete the context after you have
	// finished using the VDB.
	int nSize = contexts.getSize();
	for( int n = 0; n < nSize; ++n )
	{
		hkpPhysicsContext* context = (hkpPhysicsContext*)contexts[n];
		context->removeReference();
	}
	g_world->unlock();
#endif
}

void UpdateVisualDebuger()
{
#ifdef USE_VDB
	hkVisualDebugger* vdb = g_vdb;
	// Step the visual debugger. We first synchronize the timer data
	if( g_threadPool )
	{
		int nSize = g_contexts.getSize();
		for( int n = 0; n < nSize; ++n )
		{
			hkpPhysicsContext* context = (hkpPhysicsContext*)g_contexts[n];
			context->syncTimers( g_threadPool );
		}
	}
	g_contexts.clearAndDeallocate();
	vdb->step();
#endif
}


#ifdef USE_MULTITHREAD

bool InitHavok()
{
	hkMemoryRouter* memoryRouter = hkMemoryInitUtil::initDefault(hkMallocAllocator::m_defaultMallocAllocator, hkMemorySystem::FrameInfo(500000));
	hkBaseSystem::init( memoryRouter, HavokErrorReportFunction );

	// Get the number of physical threads available on the system
	hkHardwareInfo hwInfo;
	hkGetHardwareInfo(hwInfo);
	int totalNumThreadsUsed = hwInfo.m_numThreads;
	int g_totalNumThreadsUsed = totalNumThreadsUsed;

	// We use one less than this for our thread pool, because we must also use this thread for our simulation
	hkCpuJobThreadPoolCinfo threadPoolCinfo;
	threadPoolCinfo.m_numThreads = totalNumThreadsUsed - 1;

	// This line enables timers collection, by allocating 200 Kb per thread.  If you leave this at its default (0),
	// timer collection will not be enabled.
	threadPoolCinfo.m_timerBufferPerThreadAllocation = 200000;
	g_threadPool = new hkCpuJobThreadPool( threadPoolCinfo );

	// We also need to create a Job queue. This job queue will be used by all Havok modules to run multithreaded work.
	// Here we only use it for physics.
	hkJobQueueCinfo info;
	info.m_jobQueueHwSetup.m_numCpuThreads = totalNumThreadsUsed;
	g_jobQueue = new hkJobQueue(info);

	//
	// Enable monitors for this thread.
	//

	// Monitors have been enabled for thread pool threads already (see above comment).
	hkMonitorStream::getInstance().resize(200000);


	//
	// <PHYSICS-ONLY>: Create the physics world.
	// At this point you would initialize any other Havok modules you are using.
	//
	hkpWorld* physicsWorld;
	{
		// The world cinfo contains global simulation parameters, including gravity, solver settings etc.
		hkpWorldCinfo worldInfo;

		// Set the simulation type of the world to multi-threaded.
		worldInfo.m_simulationType = hkpWorldCinfo::SIMULATION_TYPE_MULTITHREADED;

		// Flag objects that fall "out of the world" to be automatically removed - just necessary for this physics scene
		worldInfo.m_broadPhaseBorderBehaviour = hkpWorldCinfo::BROADPHASE_BORDER_REMOVE_ENTITY;

		worldInfo.setBroadPhaseWorldSize( 10000.0f );

		physicsWorld = new hkpWorld(worldInfo);

		// Disable deactivation, so that you can view timers in the VDB. This should not be done in your game.
		physicsWorld->m_wantDeactivation = false;


		// When the simulation type is SIMULATION_TYPE_MULTITHREADED, in the debug build, the sdk performs checks
		// to make sure only one thread is modifying the world at once to prevent multithreaded bugs. Each thread
		// must call markForRead / markForWrite before it modifies the world to enable these checks.
		physicsWorld->markForWrite();


		// Register all collision agents, even though only box - box will be used in this particular example.
		// It's important to register collision agents before adding any entities to the world.
		hkpAgentRegisterUtil::registerAllAgents( physicsWorld->getCollisionDispatcher() );

		// We need to register all modules we will be running multi-threaded with the job queue
		physicsWorld->registerWithJobQueue( g_jobQueue );
	}

	g_world = physicsWorld;

	// A stopwatch for waiting until the real time has passed
	g_stopWatch.start();
	g_lastTime = g_stopWatch.getElapsedSeconds();

	// Replace filter
	hkpGroupFilter* groupFilter = new hkpGroupFilter();
	groupFilter->disableCollisionsBetween(LAYER_ACTOR, LAYER_ACTOR);

	physicsWorld->setCollisionFilter( groupFilter, true);
	groupFilter->removeReference();
	return true;
}


void UnInitHavok()
{
	hkpWorld* physicsWorld = g_world;
    
	// <PHYSICS-ONLY>: cleanup physics
	{
		physicsWorld->markForWrite();
		physicsWorld->removeReference();
	}

	delete g_jobQueue;

	// Clean up the thread pool
	g_threadPool->removeReference();

	hkBaseSystem::quit();
    hkMemoryInitUtil::quit();
}


void UpdateHavok( hkVector4& pos, hkVector4& t )
{
	hkpWorld* physicsWorld = g_world;

	//
	// Simulate the world for 1 minute.
	//


	// Take fixed time steps of 1/60th of a second.
	// This works well if your game runs solidly at 60Hz. If your game runs at 30Hz
	// you can take either 2 60Hz steps or 1 30Hz step. Note that at lower frequencies (i.e. 30 Hz)
	// more bullet through paper issues appear, and constraints will not be as stiff.
	// If you run at variable frame rate, or are likely to drop frames, you can consider
	// running your physics for a variable number of steps based on the system clock (i.e. last frame time).
	// Please refer to the user guide section on time stepping for a full treatment of this issue.


	hkReal timestep = 1.f / 60.f;
	// Pause until the actual time has passed
	if( g_stopWatch.getElapsedSeconds() - g_lastTime < timestep ) 
		return;

	g_lastTime = g_stopWatch.getElapsedSeconds();
	
	// <PHYSICS-ONLY>:
	// Step the physics world. This single call steps using this thread and all threads
	// in the threadPool. For other products you add jobs, call process all jobs and wait for completion.
	// See the multithreading chapter in the user guide for details 
	physicsWorld->stepMultithreaded( g_jobQueue, g_threadPool, timestep );
	

	UpdateVisualDebuger();

	// Clear accumulated timer data in this thread and all slave threads
	hkMonitorStream::getInstance().reset();
	g_threadPool->clearTimerData();

	// Step the graphics display (none in this demo).
	//ϳ Rigid body  ϴ.
	pos = g_bodies[0]->getPosition();

	const hkTransform &trans = g_bodies[0]->getTransform();
	t = trans.getTranslation();
	//const hkVector4 &t = trans.getTranslation();
}

#else
bool InitHavok()
{
	//hkMallocAllocator baseMalloc;
    hkMemoryRouter* memoryRouter = hkMemoryInitUtil::initDefault( hkMallocAllocator::m_defaultMallocAllocator, hkMemorySystem::FrameInfo(1024 * 1024) );
    hkBaseSystem::init(memoryRouter, HavokErrorReportFunction);

    hkpWorldCinfo info;
    info.m_simulationType = hkpWorldCinfo::SIMULATION_TYPE_CONTINUOUS;
    info.m_collisionTolerance = 0.001f; 
    info.setBroadPhaseWorldSize( 10000.0f );
    info.setupSolverInfo( hkpWorldCinfo::SOLVER_TYPE_8ITERS_HARD );
    info.m_gravity = hkVector4(0.0f, -9.8f, 0.0f);
    g_world = new hkpWorld(info);

    hkpAgentRegisterUtil::registerAllAgents( g_world->getCollisionDispatcher() );

	return true;
}

void UnInitHavok()
{
	g_world->markForWrite();
	g_world->removeReference();
    //delete g_world;
    hkBaseSystem::quit();
    hkMemoryInitUtil::quit();
}


void UpdateHavok( hkVector4& pos, hkVector4& t )
{
	g_world->stepDeltaTime(1.0f/60.0f);

	//ϳ Rigid body  ϴ.
	pos = g_bodies[0]->getPosition();

	const hkTransform &trans = g_bodies[0]->getTransform();
	t = trans.getTranslation();
	//const hkVector4 &t = trans.getTranslation();

	UpdateVisualDebuger();
}

#endif


bool MakeGround()
{
    const hkVector4 halfExtents(2000.0f, g_groundHalfHeight, 2000.0f);
    hkpBoxShape* groundShape = new hkpBoxShape(halfExtents);

    hkpRigidBodyCinfo bodyInfo;
    bodyInfo.m_mass = 0.0f; 
    bodyInfo.m_shape = groundShape;
    bodyInfo.m_motionType = hkpMotion::MOTION_FIXED;
    bodyInfo.m_position.set(0.0f, -g_groundHalfHeight, 0.0f);
	bodyInfo.m_collisionFilterInfo = LAYER_STATIC_BACKGROUND;

    g_groundBody = new hkpRigidBody(bodyInfo);
    groundShape->removeReference();

    g_world->addEntity( g_groundBody );
    g_groundBody->removeReference(); 
	return true;
}

bool MakeBox()
{
    const hkVector4 halfExtents(g_bx, g_by, g_bz);
    hkpShape* shape = new hkpBoxShape(halfExtents, 0.0f);

    hkpMassProperties massProperties;
    hkpInertiaTensorComputer::computeShapeVolumeMassProperties(shape, 5.0f, massProperties);

    hkpRigidBodyCinfo bodyInfo;
    bodyInfo.m_mass = massProperties.m_mass;
    bodyInfo.m_centerOfMass = massProperties.m_centerOfMass;
    bodyInfo.m_inertiaTensor = massProperties.m_inertiaTensor;
    bodyInfo.m_shape = shape;
    bodyInfo.m_motionType = hkpMotion::MOTION_BOX_INERTIA;
	bodyInfo.m_collisionFilterInfo = LAYER_ACTOR;

    int i = 0;
    for(int x = 0; x < g_nx; x++)
    for(int y = 0; y < g_ny; y++)
    for(int z = 0; z < g_nz; z++) {
        hkpRigidBody *body = new hkpRigidBody(bodyInfo);
        body->setPosition(hkVector4(x * g_bx * 2, 100.0f + y * g_by * 2, z * g_bz * 2));
        g_bodies[i] = body;
        g_world->addEntity(body);
        body->removeReference();
        i++;
    }
    shape->removeReference();

	return true;
}




